@@ -18,10 +18,11 @@ from __future__ import absolute_import
 
 import contextlib
 from distutils import version
+import functools
 import logging
 import string
+import threading
 
-from concurrent import futures
 from oslo_utils import encodeutils
 from oslo_utils import strutils
 import redis
@@ -45,19 +46,22 @@ def _translate_failures():
     try:
         yield
     except (exceptions.ConnectionError, exceptions.TimeoutError) as e:
-        coordination.raise_with_cause(coordination.ToozConnectionError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
     except exceptions.RedisError as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
 
 class RedisLock(locking.Lock):
     def __init__(self, coord, client, name, timeout):
         name = "%s_%s_lock" % (coord.namespace, six.text_type(name))
         super(RedisLock, self).__init__(name)
+        # NOTE(jd) Make sure we don't release and heartbeat at the same time by
+        # using a exclusive access lock (LP#1557593)
+        self._exclusive_access = threading.Lock()
         self._lock = client.lock(name,
                                  timeout=timeout,
                                  thread_local=False)
@@ -76,37 +80,47 @@ class RedisLock(locking.Lock):
         with _translate_failures():
             return bool(self._client.delete(self.name))
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
         blocking, timeout = utils.convert_blocking(blocking)
         with _translate_failures():
             acquired = self._lock.acquire(
                 blocking=blocking, blocking_timeout=timeout)
             if acquired:
-                self._coord._acquired_locks.add(self)
+                with self._exclusive_access:
+                    self._coord._acquired_locks.add(self)
             return acquired
 
     def release(self):
-        if not self.acquired:
-            return False
-        with _translate_failures():
-            try:
-                self._lock.release()
-            except exceptions.LockError:
+        with self._exclusive_access:
+            if not self.acquired:
                 return False
-            self._coord._acquired_locks.discard(self)
-            return True
+            with _translate_failures():
+                try:
+                    self._lock.release()
+                except exceptions.LockError as e:
+                    LOG.error("Unable to release lock '%r': %s", self, e)
+                    return False
+                finally:
+                    self._coord._acquired_locks.discard(self)
+                return True
 
     def heartbeat(self):
-        if self.acquired:
-            with _translate_failures():
-                self._lock.extend(self._lock.timeout)
+        with self._exclusive_access:
+            if self.acquired:
+                with _translate_failures():
+                    self._lock.extend(self._lock.timeout)
+                    return True
+        return False
 
     @property
     def acquired(self):
         return self in self._coord._acquired_locks
 
 
-class RedisDriver(coordination.CoordinationDriverCachedRunWatchers):
+class RedisDriver(coordination.CoordinationDriverCachedRunWatchers,
+                  coordination.CoordinationDriverWithExecutor):
     """Redis provides a few nice benefits that act as a poormans zookeeper.
 
     It **is** fully functional and implements all of the coordination
@@ -124,6 +138,14 @@ class RedisDriver(coordination.CoordinationDriverCachedRunWatchers):
       some notion of HA (values *can* be lost when a failover transition
       occurs).
 
+    The Redis driver connection URI should look like::
+
+      redis://[:PASSWORD@]HOST:PORT[?OPTION=VALUE[&OPTION2=VALUE2[&...]]]
+
+    For a list of options recognized by this driver, see the documentation
+    for the member CLIENT_ARGS, and to determine the expected types of those
+    options see CLIENT_BOOL_ARGS, CLIENT_INT_ARGS, and CLIENT_LIST_ARGS.
+
     To use a `sentinel`_ the connection URI must point to the sentinel server.
     At connection time the sentinel will be asked for the current IP and port
     of the master and then connect there. The connection URI for sentinel
@@ -314,26 +336,25 @@ return 1
     .. _Lua: http://www.lua.org
     """
 
+    EXCLUDE_OPTIONS = CLIENT_LIST_ARGS
+
     def __init__(self, member_id, parsed_url, options):
-        super(RedisDriver, self).__init__(member_id)
-        options = utils.collapse(options, exclude=self.CLIENT_LIST_ARGS)
+        super(RedisDriver, self).__init__(member_id, parsed_url, options)
         self._parsed_url = parsed_url
-        self._options = options
-        self._encoding = options.get('encoding', self.DEFAULT_ENCODING)
-        timeout = options.get('timeout', self.CLIENT_DEFAULT_SOCKET_TO)
+        self._encoding = self._options.get('encoding', self.DEFAULT_ENCODING)
+        timeout = self._options.get('timeout', self.CLIENT_DEFAULT_SOCKET_TO)
         self.timeout = int(timeout)
-        self.membership_timeout = float(options.get(
+        self.membership_timeout = float(self._options.get(
             'membership_timeout', timeout))
-        lock_timeout = options.get('lock_timeout', self.timeout)
+        lock_timeout = self._options.get('lock_timeout', self.timeout)
         self.lock_timeout = int(lock_timeout)
-        namespace = options.get('namespace', self.DEFAULT_NAMESPACE)
+        namespace = self._options.get('namespace', self.DEFAULT_NAMESPACE)
         self._namespace = utils.to_binary(namespace, encoding=self._encoding)
         self._group_prefix = self._namespace + b"_group"
         self._beat_prefix = self._namespace + b"_beats"
         self._groups = self._namespace + b"_groups"
         self._client = None
         self._acquired_locks = set()
-        self._executor = utils.ProxyExecutor.build("Redis", options)
         self._started = False
         self._server_info = {}
         self._scripts = {}
@@ -420,14 +441,14 @@ return 1
         return redis.StrictRedis(**kwargs)
 
     def _start(self):
-        self._executor.start()
+        super(RedisDriver, self)._start()
         try:
             self._client = self._make_client(self._parsed_url, self._options,
                                              self.timeout)
         except exceptions.RedisError as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             # Ensure that the server is alive and not dead, this does not
             # ensure the server will always be alive, but does insure that it
@@ -501,7 +522,7 @@ return 1
         for lock in self._acquired_locks.copy():
             try:
                 lock.heartbeat()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to heartbeat lock '%s'", lock,
                             exc_info=True)
         return min(self.lock_timeout, self.membership_timeout)
@@ -511,9 +532,9 @@ return 1
             lock = self._acquired_locks.pop()
             try:
                 lock.release()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to release lock '%s'", lock, exc_info=True)
-        self._executor.stop()
+        super(RedisDriver, self)._stop()
         if self._client is not None:
             # Make sure we no longer exist...
             beat_id = self._encode_beat_id(self._member_id)
@@ -522,7 +543,7 @@ return 1
                 # exist in the first place, which is fine/expected/desired...
                 with _translate_failures():
                     self._client.delete(beat_id)
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to delete heartbeat key '%s'", beat_id,
                             exc_info=True)
             self._client = None
@@ -532,14 +553,14 @@ return 1
 
     def _submit(self, cb, *args, **kwargs):
         if not self._started:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
         return self._executor.submit(cb, *args, **kwargs)
 
     def _get_script(self, script_key):
         try:
             return self._scripts[script_key]
         except KeyError:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
 
     def create_group(self, group_id):
         script = self._get_script('create_group')
@@ -696,13 +717,13 @@ return 1
             if result == -3:
                 raise coordination.GroupNotEmpty(group_id)
             if result == -4:
-                raise coordination.ToozError("Unable to remove '%s' key"
-                                             " from set located at '%s'"
-                                             % (args[0], keys[-1]))
+                raise tooz.ToozError("Unable to remove '%s' key"
+                                     " from set located at '%s'"
+                                     % (args[0], keys[-1]))
             if result != 1:
-                raise coordination.ToozError("Internal error, unable"
-                                             " to complete group '%s' removal"
-                                             % (group_id))
+                raise tooz.ToozError("Internal error, unable"
+                                     " to complete group '%s' removal"
+                                     % (group_id))
 
         return RedisFutureResult(self._submit(_delete_group, script))
 
@@ -738,24 +759,5 @@ return 1
         return result
 
 
-class RedisFutureResult(coordination.CoordAsyncResult):
-    """Redis asynchronous result that references a future."""
-
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=10):
-        try:
-            # Late translate the common failures since the redis client
-            # may throw things that we can not catch in the callbacks where
-            # it is used (especially one that uses the transaction
-            # method).
-            with _translate_failures():
-                return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
-
-    def done(self):
-        return self._fut.done()
+RedisFutureResult = functools.partial(coordination.CoordinatorResult,
+                                      failure_translator=_translate_failures)
